{-# OPTIONS -XRank2Types #-}

{-| Class of monad transformers.
-}

module Monatron.MonadT (
  MonadT(..), 
  {-| The following should hold:
    
    * 'lift' is a monad morphism
    
    >         treturn = return . lift
    >   lift (m >= f) = lift m `tbind` lift . f

    * @(t m,'treturn','tbind')@ is a monad
    
    >              treturn a `tbind` f = f a
    >                m `tbind` treturn = m
    >   m `tbind` (\a -> f a \tbind\ g = (m `tbind` f) `tbind` g

    * 'tmixmap' is functorial
    
    >                 tmixmap id id = id
    >   tmixmap f g' . tmixmap f' g = tmixmap (f' . f) (g' . g)
    
    -}
  tmap,
  module Control.Monad
) where
 
import Control.Monad

----------------------------------------------------------
-- Monad transformer class
----------------------------------------------------------

-- | A complete instance should provide implementations of 'lift', tbind', 
-- and 'tmixmap'.
class MonadT t where
    -- | Lifts a computation to the transformed monad.
    lift    :: Monad m  => m a -> t m a
    -- | @return@ of the transformed monad
    treturn :: Monad m => a -> t m a
    treturn =  lift. return
    -- | bind @(>>=)@ for the transformed monad
    tbind   :: Monad m => t m a -> (a -> t m b) -> t m b
    -- | maps the first function to the covariant occurrences of @m@,
    -- and the second to the contravariant occurrences.
    tmixmap :: (Monad m, Monad n ) => 
               (forall a.   m a ->   n a) -> 
               (forall b.   n b ->   m b) -> 
                          t m c -> t n c

-- | Applies the argument function to every occurence of the monad inside 
--   the transformer.
tmap   :: (Monad m, MonadT t)  => (forall x. m x -> m x) -> t m a -> t m a
tmap m = tmixmap m m
